from warnings import warn
import iRRAPy.RRAresult as rres
import iRRAPy.Border as bod
import iRRAPy.errors as err
import iRRAPy.utilRRAcost as utilCost
import iRRAPy.utilRRAgeometry as geo
import iRRAPy.utilRRAareasROC as rocae
import iRRAPy.utilROCgeometry as rocgeo
import iRRAPy.utilRRAareas as ae
from warnings import simplefilter

simplefilter("always", UserWarning)

"""Computes the RRA of a ROC curve. It's possible to use any performance metric
to build the RoI. It returns the total area of the RoI, the area of the RoI
under the curve and all the metrics used to generate it. print and plot allow
the user to print the resulting RRA and to plot the ROC curve and the RoI"""

def rra(roc_x, roc_y, AP, AN,
        *,precision = False, c_precision = "pop", p_precision = -1,
        recall = False, c_recall = "pop", p_recall = -1,
        fm = False, c_fm = "pop", p_fm = -1,
        npv = False, c_npv = "pop", p_npv = -1,
        specificity = False, c_specificity = "pop", p_specificity = -1,
        fallout = False, c_fallout = "pop", p_fallout = -1,
        nm = False, c_nm = "pop", p_nm = -1,
        j = False, c_j = -1,
        markedness = False, c_markedness = -1,
        phi = False, c_phi = 0.4,
        ncost = False, c_ncost = "uses_mu", lamb = [-1], mu = 1,
        plot = True, print = True,
        cROI = "grey", cUnder = "c", col="k", 
        xlab = "FP/AN", ylab = "TP/AP", main = "ROC curve and RoI", 
        xlim = [0,1], ylim = [0,1], asp = 'auto',
        ls = "-", lw = 1, ADD = False, diagonal = True, show = True,
        label = "ROC curve"):
    n = AP+AN
    """This will order the input so it can be used without any problem."""
    if roc_x[0] == 1:
        roc_x.reverse()
        roc_y.reverse()
    """ input check """
    if len(roc_x) != len(roc_y):
        raise err.InputValueError(("Arguments \"roc_x\" and \"roc_y\" must"
                               " have the same length"))
    if AP<=0 or AN<0:
        raise err.InputValueError(("AP and AN must be greater than zero."
                              " AN can be equal to zero"))
    for x in roc_x:
        if x < 0 or x > 1:
            raise err.InputValueError(("\"roc_x\" values must be"
                                            " between 0 and 1"))
    for y in roc_y:
        if y < 0 or y > 1:
            raise err.InputValueError(("\"roc_y\" values must be"
                                            " between 0 and 1"))
    if n < len(roc_x):
        warn(("\nThe number of thresholds considered by the ROC curve"
                "is greater than the number of observations (AP+AN)"))

    """ end input check """
    if not(precision or recall or fm or npv or specificity or fallout or nm or
           j or markedness or phi or ncost):
        warn(("\nNo performance metrics have been selected."
                " The AUC value of the ROC curve has been returned"))
        auc = rocae.roc_curve_auc(roc_x, roc_y)
        result = rres.RRAresult(roc_x, roc_y, [0,1,1,0], [0,0,1,1], 
                        [[*roc_x, 1]], [[*roc_y, 0]], 1, auc, [], [])
    else:
        rra_is_zero = False
        k = AN/AP
        auc = rocae.roc_curve_auc(roc_x, roc_y)
        borders = []

        """ METRICS CHECK AND BORDER GENERATION """
        if(precision):
            tmp = []
            tmp.append("precision")
            if c_precision == "uni":
                tmp.append(("p(m) = " + str(p_precision)))
            else:
                tmp.append(c_precision)
            bt = bod.precision_border(c_precision, p_precision, k)
            tmp.append(bt.slope)
            tmp.append(bt.intercept)
            borders.append(tmp)
            if not rra_is_zero:
                rra_is_zero = bt.rra_zero

        if(recall):
            tmp = []
            tmp.append("recall")
            if c_recall == "uni":
                tmp.append(("p(m) = " + str(p_recall)))
            else:
                tmp.append(c_recall)
            bt = bod.recall_border(c_recall, p_recall, k)
            tmp.append(bt.slope)
            tmp.append(bt.intercept)
            borders.append(tmp)
            if not rra_is_zero:
                rra_is_zero = bt.rra_zero

        if(fm):
            tmp = []
            tmp.append("FM")
            if c_fm == "uni":
                tmp.append(("p(m) = " + str(p_fm)))
            else:
                tmp.append(c_fm)
            bt = bod.fm_border(c_fm, p_fm, k)
            tmp.append(bt.slope)
            tmp.append(bt.intercept)
            borders.append(tmp)
            if not rra_is_zero:
                rra_is_zero = bt.rra_zero

        if(npv):
            tmp = []
            tmp.append("NPV")
            if c_npv == "uni":
                tmp.append(("p(m) = " + str(p_npv)))
            else:
                tmp.append(c_npv)
            bt = bod.npv_border(c_npv, p_npv, k)
            tmp.append(bt.slope)
            tmp.append(bt.intercept)
            borders.append(tmp)
            if not rra_is_zero:
                rra_is_zero = bt.rra_zero

        if(specificity):
            tmp = []
            tmp.append("specificity")
            if c_specificity == "uni":
                tmp.append(("p(m) = " + str(p_specificity)))
            else:
                tmp.append(c_specificity)
            bt = bod.specificity_border(c_specificity, p_specificity, k)
            tmp.append(bt.slope)
            tmp.append(bt.intercept)
            borders.append(tmp)
            if not rra_is_zero:
                rra_is_zero = bt.rra_zero

        if(fallout):
            tmp = []
            tmp.append("fallout")
            if c_fallout == "uni":
                tmp.append(("p(m) = " + str(p_fallout)))
            else:
                tmp.append(c_fallout)
            bt = bod.fallout_border(c_fallout, p_fallout, k)
            tmp.append(bt.slope)
            tmp.append(bt.intercept)
            borders.append(tmp)
            if not rra_is_zero:
                rra_is_zero = bt.rra_zero

        if(nm):
            tmp = []
            tmp.append("NM")
            if c_nm == "uni":
                tmp.append(("p(m) = " + str(p_nm)))
            else:
                tmp.append(c_nm)
            bt = bod.nm_border(c_nm, p_nm, k)
            tmp.append(bt.slope)
            tmp.append(bt.intercept)
            borders.append(tmp)
            if not rra_is_zero:
                rra_is_zero = bt.rra_zero

        if(j):
            tmp = []
            tmp.append("J")
            tmp.append(c_j)
            bt = bod.j_border(c_j)
            tmp.append(bt.slope)
            tmp.append(bt.intercept)
            borders.append(tmp)
            if not rra_is_zero:
                rra_is_zero = bt.rra_zero

        if(ncost):
            NC = k/((1+k)**2)
            if ((isinstance(c_ncost, str) and c_ncost != "uses_mu") or
                (isinstance(c_ncost, (int,float)) and (c_ncost <= 0 or
                    c_ncost > (NC)))):
                raise err.InputValueError(("\"c_ncost\" value must be between"
                                           " 0 and k/((1+k)^2) with k = AN/AP"
                                           " (0 escluded) or \"uses_mu\""))
            if isinstance(c_ncost, (int,float)):
                mu = utilCost.compute_by_ref(c_ncost, k)
            if ((len(lamb) == 1) or (len(lamb) == 2 and lamb[0] == lamb[1])):
                tmp = []
                tmp.append("NC")
                tmp.append(("a maximum NC value (mu*NC(pop))" + 
                            " with lambda =" + str(lamb[0]) + 
                            "and mu = " + str(mu)))
                bt = bod.ncost_border(lamb[0], mu, k)
                tmp.append(bt.slope)
                tmp.append(bt.intercept)
                borders.append(tmp)
                if not rra_is_zero:
                    rra_is_zero = bt.rra_zero
            elif lamb[0] < lamb[1]:
                tmp = []
                tmp.append("NC1")
                tmp.append(("lambda between " + str(lamb[0]) + " and " +
                            str(lamb[1]) + "; mu = " + str(mu)))
                bt = bod.ncost_border(lamb[0], mu, k)
                tmp.append(bt.slope)
                tmp.append(bt.intercept)
                borders.append(tmp)
                if not rra_is_zero:
                    rra_is_zero = bt.rra_zero

                tmp = []
                tmp.append("NC2")
                tmp.append(("lambda between " + str(lamb[0]) + " and " +
                            str(lamb[1]) + "; mu = " + str(mu)))
                bt = bod.ncost_border(lamb[1], mu, k)
                tmp.append(bt.slope)
                tmp.append(bt.intercept)
                borders.append(tmp)
                if not rra_is_zero:
                    rra_is_zero = bt.rra_zero
            else:
                raise err.InputValueError(("lambda represents a range of"
                                           " values. Therefore, the first"
                                           " value must be lesser or equal to"
                                           " the second"))
        if(phi):
            tmp = []
            if c_phi < 0 or c_phi > 1:
                raise err.InputValueError("phi value must be between 0 and 1")
            if c_phi == 1:
                warn(("\nThe selected phi represents a non existent RoI."
                    " Unless a perfect ROC curve is being considered, the"
                    " resulting RRA will be equal to zero"))
                rra_is_zero = True
            elif k == 0:
                warn(("\nAN/AP is equal to zero. Therefore, the phi border"
                    " represents a non existent RoI. Unless a perfect ROC"
                    " curve is being considered, the resulting RRA will be"
                    " equal to zero"))
                rra_is_zero = True
            elif c_phi == 0:
                tmp.append("phi")
                tmp.append(c_phi)
                tmp.append(1)
                tmp.append(0)
            else:
                tmp.append("phi")
                tmp.append(c_phi)
                tmp.append(float("inf"))
                tmp.append(geo.intersect_phi_vertical_line(c_phi, k, 0))
            borders.append(tmp)
            
        if(markedness):
            tmp = []
            if c_markedness < 0 or c_markedness > 1:
                raise err.InputValueError(("markedness value must be between"
                                            " 0 and 1"))
            tmp.append("markedness")
            tmp.append(c_markedness)
            if c_markedness == 1:
                warn(("\nThe selected markedness value represents a non"
                    " existent RoI. Unless a perfect ROC curve is being"
                    " considered, the resulting RRA will be equal to zero"))
                rra_is_zero = True
            elif k == 0:
                warn(("\nAN/AP is equal to zero. Therefore, the markedness"
                    " border represents a non existent RoI. Unless a perfect"
                    " ROC curve is being considered, the resulting RRA will"
                    " be equal to zero"))
                rra_is_zero = True
            elif c_markedness == 0:
                tmp.append(1)
                tmp.append(0)
            else:
                tmp.append(float("inf"))
                tmp.append(geo.intersect_markedness_vertical_line(c_markedness,
                                                                    k, 0))
            borders.append(tmp)
            
        """END performance metrics check"""
        """rra_is_zero means that the RoI is non-existent because it is
        established by a point (i.e. (0,1)) or by a line (i.e. y=1 or x=0).
        The resulting RRA is zero unless the ROC curve is perfect (AUC = 1)"""
        if rra_is_zero:
            if auc == 1:
                result = rres.RRAresult(roc_x,roc_y,[],[],[],[],1,1,[],[])
            else:
                result = rres.RRAresult(roc_x,roc_y,[],[],[],[],1,0,[],[])
        else:
            borders.sort(key = lambda x: x[2], reverse = True)
            borders.sort(key = lambda x: x[3], reverse = True)
            coords = generate_borders(roc_x, roc_y, k, c_phi, 
                                                c_markedness, borders)
            result = rres.RRAresult(roc_x, roc_y, coords[0], coords[1],
                                    coords[2], coords[3], coords[4], 
                                    coords[5], [e[0] for e in borders], 
                                    [e[1] for e in borders])
    if print:
        result.print()
    if plot:
        result.plot(cROI,cUnder,col,xlab,ylab,main,xlim,ylim,asp,ls,lw,
                                                    ADD,diagonal,show,label)
        
    return result

"""This function returns the coordinates of the RoI and its area. It also 
returns the area of the RoI under the curve and ita area"""
def generate_borders(roc_x, roc_y, k, c_phi, c_markedness, borders):
    met = borders[0][0]
    startind = -1
    xx = -1
    if (met == "fallout" or met == "specificity" or ((met == "NC" or 
        met == "NC1" or met == "NC2") and borders[0][2] == -1)):
        if len(borders) == 1:
            xx = borders[0][3]
        else:
            met = borders[1][0]
            if (met == "fallout" or met == "specificity" or ((met == "NC" or 
                met == "NC1" or met == "NC2") and borders[1][2] == -1)):
                if len(borders) == 2:
                    xx = borders[1][3]
                else:
                    startind = 2
            else:
                startind = 1
        if xx != -1:
            area = rocae.roc_curve_vertical_line(roc_x, roc_y, xx)
            roi = xx
            procc = rocgeo.partial_roc_curve(roc_x, roc_y, 0, xx)
            result = [[0,xx,xx,0], [0,0,1,1], [[*procc[0],xx,0]], 
                    [[*procc[1],0,0]], roi, area]
    else:
        startind = 0
    
    if startind != -1:
        big_area = 0
        area_under = 0
        b_interx = [0]
        b_intery = [borders[startind][3]]
        if b_intery[0] < 0:
            intersection = geo.intersect_lines(0,0,borders[startind][2],
                                                borders[startind][3])
            b_interx.append(intersection.x)
            b_intery = [0,0]
            big_area += intersection.x
            area_under += rocae.roc_curve_vertical_line(roc_x, roc_y,
                                                        intersection.x)
        roc_interx = []
        roc_intery = []
        """When two metrics intersect x=0 in the same point, the one
        with a higher slope is selected. This is already defined by the
        ordered list "borders". If a curve is considered instead,
        it depends on whether they intersect the line in x>0. If the two
        curve, phi and markedness, are considered, markedness is always
        the choice"""
        if (len(borders) > startind+1 and 
                            borders[startind+1][3] == b_intery[0]):
            if (borders[startind][2] == float("inf") and 
                                borders[startind+1][2] == float("inf")):
                if borders[startind+1][0] == "markedness":
                    startind += 1
        if (len(borders) > startind+1 and 
                    borders[startind+1][3] == b_intery[0] and 
                        (borders[startind+1][0] != "fallout" or 
                            borders[startind+1][0] != "specificity")):
            if borders[startind][2] == float("inf"):
                if borders[startind][0] == "phi":
                    intersection = geo.intersect_phi_line(c_phi,k,
                                            borders[startind+1][2],
                                            borders[startind+1][3])
                else:
                    intersection= geo.intersect_markedness_line(c_markedness,
                                                    k,borders[startind+1][2],
                                                    borders[startind+1][3])
                if len(intersection) == 1:
                    startind += 1
        
        i = startind
        ind = i
        done = False
        metf = ""
        slopef = 0
        interceptf = 0
        
        while not done:
            j = (i+1)%len(borders)
            xcurrent = b_interx[-1]
            metf = borders[i][0]
            slopef = borders[i][2]
            interceptf = borders[i][3]
            
            xmin = 2
            ymin = 2
            slopemin = 0
            metmin = ""
            interceptmin = 0
            
            """First, intersection with y=1"""
            if slopef != 0:
                if slopef == float("inf"):
                    if metf == "phi":
                        intersection = geo.intersect_phi_line(c_phi,k,0,1)
                        xmin = min(e.x for e in intersection)
                    else:
                        intersection=geo.intersect_markedness_line(
                                                        c_markedness,k,0,1)
                        xmin = min(e.x for e in intersection)
                elif slopef != -1:
                    intersection = geo.intersect_lines(slopef, interceptf,0,1)
                    xmin = intersection.x
                else:
                    xmin = interceptf
                slopemin = 0
                interceptmin = 1
                ymin = 1
                
            """Then, if xmin > 1, x=1"""
            if slopef != -1 and xmin > 1:
                xmin = 1
                slopemin = -1
                interceptmin = 1
                if slopef == float("inf"):
                    if metf == "phi":
                        ymin = geo.intersect_phi_vertical_line(c_phi,k,1)
                    else:
                        ymin = geo.intersect_markedness_vertical_line(
                                                        c_markedness,k,1)
                else:
                    ymin=geo.intersect_line_vertical_line(slopef,interceptf,1)
            
            """It will intersect the current piece of border with the others
            The best intersection point (minimum xmin and maximum ymin) will
            become one of the vertices of the RoI"""
            if slopef != -1:
                while j != i:
                    metl = borders[j][0]
                    slopel = borders[j][2]
                    interceptl = borders[j][3]
                    xselect = 0
                    yselect = 0
                    if slopef == float("inf"):
                        if slopel == float("inf"):
                            inter = geo.intersect_phi_markedness(c_phi,
                                                            c_markedness,k)
                        else:
                            if slopel != -1:
                                if metf == "phi":
                                    inter = geo.intersect_phi_line(
                                            c_phi, k, slopel, interceptl)
                                else:
                                    inter = geo.intersect_markedness_line(
                                            c_markedness,k,slopel,interceptl)
                            elif interceptl > xcurrent:
                                x1 = interceptl
                                if metf == "phi":
                                    y1=geo.intersect_phi_vertical_line(c_phi,
                                                                        k,x1)
                                else:
                                    y1=geo.intersect_markedness_vertical_line(
                                                        c_markedness,k,x1)
                                inter = [geo.Intersection(x1,y1)]
                    else:
                        if slopel == float("inf"):
                            if metl == "phi":
                                inter = geo.intersect_phi_line(c_phi, k,
                                                    slopef, interceptf)
                            else:
                                inter = geo.intersect_markedness_line(
                                        c_markedness,k,slopef,interceptf)
                        else:
                            if slopel != -1:
                                inter = geo.intersect_lines(slopef,
                                interceptf,slopel,interceptl)
                                if isinstance(inter, geo.Intersection):
                                    inter = [inter]
                            elif interceptl > xcurrent:
                                x1 = interceptl
                                y1 = geo.intersect_line_vertical_line(slopef,
                                                            interceptf, x1)
                                inter = [geo.Intersection(x1,y1)]
                    if len(inter) == 0:
                        j = (j+1)%len(borders)
                        continue
                    elif len(inter) == 1:
                        x1 = inter[0].x
                        y1 = inter[0].y
                        x2 = -1
                        y2 = -1
                    elif len(inter) == 2:
                        x1 = min(inter[0].x,inter[1].x)
                        y1 = min(inter[0].y,inter[1].y)
                        x2 = max(inter[0].x,inter[1].x)
                        y2 = max(inter[0].y,inter[1].y)
                    if x1 > xcurrent:
                        xselect = x1
                        yselect = y1
                    elif x2 > xcurrent:
                        xselect = x2
                        yselect = y2
                    
                    
                    if xselect > xcurrent:
                        if (xselect < xmin or (xselect == xmin and 
                                                    yselect > ymin)):
                            xmin = xselect
                            ymin = yselect
                            ind = j
                            slopemin = slopel
                            interceptmin = interceptl
                            metmin = metl
                        elif (xselect == xmin and yselect == ymin and
                                                        slopemin != -1):
                            if slopel == -1:
                                ind = j
                                slopemin = slopel
                                interceptmin = interceptl
                                metmin = metl
                            elif (slopemin == float("inf") and 
                            slopel == float("inf") and metl == "markedness"):
                                ind = j
                                slopemin = slopel
                                interceptmin = interceptl
                                metmin = metl
                            elif (slopemin == float("inf") and 
                                                    slopel != float("inf")):
                                if metmin == "phi":
                                    inter = geo.intersect_phi_line(c_phi,
                                                        k,slopel,interceptl)
                                else:
                                    inter = geo.intersect_markedness_line(
                                            c_markedness,k,slopel,interceptl)
                                if max(inter, key = lambda e: e.x) <= xmin:
                                    ind = j
                                    slopemin = slopel
                                    interceptmin = interceptl
                                    metmin = metl
                            elif (slopemin != float("inf") and 
                                                    slopel == float("inf")):
                                if metl == "phi":
                                    inter = geo.intersect_phi_line(c_phi,
                                                    k,slopemin,interceptmin)
                                else:
                                    inter = geo.intersect_markedness_line(
                                        c_markedness,k,slopemin,interceptmin)
                                if max(inter, key = lambda e: e.x) > xmin:
                                    ind = j
                                    slopemin = slopel
                                    interceptmin = interceptl
                                    metmin = metl
                            elif (slopel > slopemin) and slopemin != -1:
                                ind = j
                                slopemin = slopel
                                interceptmin = interceptl
                                metmin = metl
                    j = (j+1)%len(borders)
                    
            xff = b_interx[-1]
            if slopef == float("inf"):
                tmp = [*range(101)]
                if metf == "phi":
                    line = bod.phi_border(c_phi, k,[i/100 for i in tmp])
                    where_y_one = geo.intersect_phi_line(c_phi,k,0,1)
                    phi_areas = rocae.roc_partial_area_phi(roc_x, roc_y,
                                                            c_phi,k,xff,xmin)
                    big_area += phi_areas[0]
                    area_under += phi_areas[1]
                    roc_inter = rocgeo.intersect_roc_curve_phi(roc_x,
                                                    roc_y,c_phi,k,xff,xmin)
                else:
                    line = bod.markedness_border(c_markedness,k,
                                                        [i/100 for i in tmp])
                    where_y_one = geo.intersect_markedness_line(
                                                        c_markedness,k,0,1)
                    mark_areas = rocae.roc_partial_area_markedness(
                                        roc_x,roc_y,c_markedness,k,xff,xmin)
                    big_area += mark_areas[0]
                    area_under += mark_areas[1]
                    roc_inter = rocgeo.intersect_roc_curve_markedness(
                                        roc_x,roc_y,c_markedness,k,xff,xmin)
                xt = min(where_y_one, key = lambda e: e.x)
                aa = [i for i,el in enumerate(line[0]) if el>xff and el<xmin]
                xx = [line[0][i] for i in aa]
                yy = [line[1][i] for i in aa]
                for pp in xx:
                    b_interx.append(pp)
                for pp in yy:
                    b_intery.append(pp)
                if done:
                    b_interx.append(xt)
                    b_intery.append(1)
                else:
                    b_interx.append(xmin)
                    b_intery.append(ymin)
            else:
                b_interx.append(xmin)
                b_intery.append(ymin)
                big_area += ae.quadrilateral_area(xff,1,b_intery[-2],
                                                b_interx[-1],1,b_intery[-1])
                area_under += rocae.roc_curve_partial_area_line(roc_x,
                                    roc_y,slopef,interceptf,xff,b_interx[-1])
                if slopef == -1:
                    roc_inter = rocgeo.intersect_roc_curve_vertical_line(
                                                    roc_x,roc_y,interceptf)
                    hh = True
                    for pp in roc_inter:
                        if pp.y >= b_intery[-2]:
                            hh = False
                            break
                    if hh:
                        roc_inter = []
                else:
                    roc_inter = rocgeo.intersect_roc_curve_line(roc_x,
                                        roc_y,slopef,interceptf,xff,xmin)
                    if interceptf < 0 and i == startind:
                        roc_inter.insert(0, geo.Intersection(0,0))
            for pp in roc_inter:
                roc_interx.append(pp.x)
                roc_intery.append(pp.y)
                
            if xmin == 1 or ymin == 1:
                done = True
            else:
                i = ind
                j = (i+1)%len(borders)
        
        roi_x = [*b_interx]
        roi_y = [*b_intery]
        if b_interx[-1] == 1:
            roi_x.append(1)
            roi_y.append(1)
        roi_x.append(0)
        roi_y.append(1)
        
        roi_under_x = []
        roi_under_y = []
        if len(roc_interx) > 0:
            if len(roc_interx)%2 != 0:
                roc_interx.append(roc_x[-1])
                roc_intery.append(roc_y[-1])
            j = 1
            while j < len(roc_interx):
                xb = roc_interx[j]
                yb = roc_intery[j]
                xa = roc_interx[j-1]
                ya = roc_intery[j-1]
                bbx = []
                bby = []
                for ii in range(len(b_interx)):
                    iii = -(ii+1)
                    if (b_interx[iii] >= xa and b_interx[iii] <= xb and
                            b_intery[iii] >= ya and b_intery[iii] <= yb):
                        bbx.append(b_interx[iii])
                        bby.append(b_intery[iii])
                rrx = []
                rry = []
                for ii in range(len(roc_x)):
                    if (roc_x[ii] >= xa and roc_x[ii] <= xb and
                            roc_y[ii] >= ya and roc_y[ii] <= yb):
                        rrx.append(roc_x[ii])
                        rry.append(roc_y[ii])
                u_x = [xb, *bbx, xa, *rrx]
                u_y = [yb, *bby, ya, *rry]
                roi_under_x = [*roi_under_x, u_x]
                roi_under_y = [*roi_under_y, u_y]
                j += 2
            
        result =[roi_x, roi_y, roi_under_x, roi_under_y, big_area, area_under]
    return result